import cirq
import numpy as np
import stim
import stimcirq


def test_repr():
    cirq.testing.assert_equivalent_repr(
        stimcirq.SweepPauli(stim_sweep_bit_index=1, pauli=cirq.X, cirq_sweep_symbol="a"),
        global_vals={"stimcirq": stimcirq},
    )
    cirq.testing.assert_equivalent_repr(
        stimcirq.SweepPauli(stim_sweep_bit_index=2, pauli=cirq.Y, cirq_sweep_symbol="b"),
        global_vals={"stimcirq": stimcirq},
    )


def test_stim_conversion():
    actual = stimcirq.cirq_circuit_to_stim_circuit(
        cirq.Circuit(
            stimcirq.SweepPauli(
                stim_sweep_bit_index=5, pauli=cirq.X, cirq_sweep_symbol="init_66"
            ).on(cirq.GridQubit(1, 3)),
            stimcirq.SweepPauli(
                stim_sweep_bit_index=7, pauli=cirq.Y, cirq_sweep_symbol="init_63"
            ).on(cirq.GridQubit(2, 4)),
            stimcirq.SweepPauli(
                stim_sweep_bit_index=5, pauli=cirq.Z, cirq_sweep_symbol="init_66"
            ).on(cirq.GridQubit(2, 4)),
        )
    )
    assert actual == stim.Circuit(
        """
        QUBIT_COORDS(1, 3) 0
        QUBIT_COORDS(2, 4) 1
        CX sweep[5] 0
        CY sweep[7] 1
        TICK
        CZ sweep[5] 1
        TICK
    """
    )

def test_resolving_without_sweep_resolver():
    cirq_circuit = cirq.Circuit(
        stimcirq.SweepPauli(
            stim_sweep_bit_index=0, pauli=cirq.X, cirq_sweep_symbol="init_0"
        ).on(cirq.LineQubit(0)),
        stimcirq.SweepPauli(
            stim_sweep_bit_index=1, pauli=cirq.X, cirq_sweep_symbol="init_1"
        ).on(cirq.LineQubit(1)),
    )
    # Resolve the cirq circuit without specifying the sweep_symbol (this should leave the op 
    # unchanged).
    resolved_circuit = cirq.resolve_parameters(cirq_circuit, {"init_1":1})
    op0, op1, = resolved_circuit.all_operations()
    assert isinstance(op0.gate, stimcirq.SweepPauli) # Operation should still be a SweepPauli
    assert isinstance(op1.gate, type(cirq.X)) # Operation should be converted to a cirq.X
    # Conversion to stim still possible
    assert stimcirq.cirq_circuit_to_stim_circuit(resolved_circuit) 

def test_parameter_names():
    cirq_circuit = cirq.Circuit(
        stimcirq.SweepPauli(
            stim_sweep_bit_index=0, pauli=cirq.X, cirq_sweep_symbol="init_0"
        ).on(cirq.LineQubit(0)),
        stimcirq.SweepPauli(
            stim_sweep_bit_index=1, pauli=cirq.X, cirq_sweep_symbol="init_1"
        ).on(cirq.LineQubit(1)),
    )
    assert cirq.parameter_names(cirq_circuit) == {'init_0', 'init_1'} 


def test_cirq_compatibility():
    circuit = cirq.Circuit(
        stimcirq.SweepPauli(stim_sweep_bit_index=5, pauli=cirq.X, cirq_sweep_symbol="xx").on(
            cirq.LineQubit(0)
        ),
        cirq.measure(cirq.LineQubit(0), key="k"),
    )
    np.testing.assert_array_equal(
        cirq.Simulator().sample(circuit, params={"xx": 0}, repetitions=3)["k"], [0, 0, 0]
    )
    np.testing.assert_array_equal(
        cirq.Simulator().sample(circuit, params={"xx": 1}, repetitions=3)["k"], [1, 1, 1]
    )

    circuit = cirq.Circuit(
        stimcirq.SweepPauli(stim_sweep_bit_index=5, pauli=cirq.Z, cirq_sweep_symbol="xx").on(
            cirq.LineQubit(0)
        ),
        cirq.measure(cirq.LineQubit(0), key="k"),
    )
    np.testing.assert_array_equal(
        cirq.Simulator().sample(circuit, params={"xx": 0}, repetitions=3)["k"], [0, 0, 0]
    )
    np.testing.assert_array_equal(
        cirq.Simulator().sample(circuit, params={"xx": 1}, repetitions=3)["k"], [0, 0, 0]
    )

    circuit = cirq.Circuit(
        cirq.H(cirq.LineQubit(0)),
        stimcirq.SweepPauli(stim_sweep_bit_index=5, pauli=cirq.Z, cirq_sweep_symbol="xx").on(
            cirq.LineQubit(0)
        ),
        cirq.H(cirq.LineQubit(0)),
        cirq.measure(cirq.LineQubit(0), key="k"),
    )
    np.testing.assert_array_equal(
        cirq.Simulator().sample(circuit, params={"xx": 0}, repetitions=3)["k"], [0, 0, 0]
    )
    np.testing.assert_array_equal(
        cirq.Simulator().sample(circuit, params={"xx": 1}, repetitions=3)["k"], [1, 1, 1]
    )


def test_json_serialization():
    c = cirq.Circuit(
        stimcirq.SweepPauli(stim_sweep_bit_index=5, pauli=cirq.X, cirq_sweep_symbol="init_66").on(
            cirq.GridQubit(1, 3)
        ),
        stimcirq.SweepPauli(stim_sweep_bit_index=7, pauli=cirq.Y, cirq_sweep_symbol="init_63").on(
            cirq.GridQubit(2, 4)
        ),
    )
    json = cirq.to_json(c)
    c2 = cirq.read_json(json_text=json, resolvers=[*cirq.DEFAULT_RESOLVERS, stimcirq.JSON_RESOLVER])
    assert c == c2


def test_json_backwards_compat_exact():
    raw = stimcirq.SweepPauli(stim_sweep_bit_index=2, pauli=cirq.Z, cirq_sweep_symbol="abc")
    packed = '{\n  "cirq_type": "SweepPauli",\n  "pauli": {\n    "cirq_type": "_PauliZ",\n    "exponent": 1.0,\n    "global_shift": 0.0\n  },\n  "stim_sweep_bit_index": 2,\n  "cirq_sweep_symbol": "abc"\n}'
    assert cirq.read_json(json_text=packed, resolvers=[*cirq.DEFAULT_RESOLVERS, stimcirq.JSON_RESOLVER]) == raw
    assert cirq.to_json(raw) == packed



from typing import Any, Dict, List, Iterable

import cirq
import stim



@cirq.value_equality
class IIErrorGate(cirq.Gate):
    """Handles explaining stim's II_ERROR gate to cirq."""

    def __init__(self, gate_args: Iterable[float]):
        self.gate_args = tuple(gate_args)

    def _num_qubits_(self) -> int:
        return 2

    def _circuit_diagram_info_(self, args: cirq.CircuitDiagramInfoArgs) -> List[str]:
        return [f'II_ERROR({self.gate_args})', f'II_ERROR({self.gate_args})']

    def _value_equality_values_(self):
        return self.gate_args

    def _decompose_(self, qubits):
        pass

    def _stim_conversion_(self, edit_circuit: stim.Circuit, targets: List[int], tag: str, **kwargs):
        edit_circuit.append('II_ERROR', targets, arg=self.gate_args, tag=tag)

    def __str__(self) -> str:
        return 'II_ERROR(' + ",".join(str(e) for e in self.gate_args) + ')'

    def __repr__(self):
        return f'stimcirq.IIErrorGate({self.gate_args!r})'

    @staticmethod
    def _json_namespace_() -> str:
        return ''

    def _json_dict_(self) -> Dict[str, Any]:
        return {'gate_args': self.gate_args}



import collections
from typing import Dict, List, Sequence

import cirq

from ._cirq_to_stim import cirq_circuit_to_stim_data


try:
    ResultImpl = cirq.ResultDict  # For cirq >= 0.14
except AttributeError:
    ResultImpl = cirq.Result  # For cirq < 0.14


class StimSampler(cirq.Sampler):
    """Samples stabilizer circuits using Stim.

    Supports circuits that contain Clifford operations, measurement operations, reset operations, and noise operations
    that can be decomposed into probabilistic Pauli operations. Unknown operations are supported as long as they provide
    a decomposition into supported operations via `cirq.decompose` (i.e. via a `_decompose_` method).

    Note that batch sampling is significantly faster (as in potentially thousands of times faster) than individual
    sampling, because it amortizes the cost of parsing and analyzing the circuit.
    """

    def run_sweep(
        self, program: cirq.Circuit, params: cirq.Sweepable, repetitions: int = 1
    ) -> Sequence[cirq.Result]:
        results: List[cirq.Result] = []
        for param_resolver in cirq.to_resolvers(params):
            # Request samples from stim.
            instance = cirq.resolve_parameters(program, param_resolver)
            converted_circuit, key_ranges = cirq_circuit_to_stim_data(instance, flatten=True)
            samples = converted_circuit.compile_sampler().sample(repetitions)

            # Find number of qubits (length), number of instances, and indices for each measurement key.
            lengths: Dict[str, int] = {}
            instances: Dict[str, int] = collections.Counter()
            indices: Dict[str, int] = collections.defaultdict(list)
            k = 0
            for key, length in key_ranges:
                prev_length = lengths.get(key)
                if prev_length is None:
                    lengths[key] = length
                elif length != prev_length:
                    raise ValueError(f"different numbers of qubits for key {key}: {prev_length} != {length}")
                instances[key] += 1
                indices[key].extend(range(k, k + length))
                k += length

            # Convert unlabelled samples into keyed results.
            records = {
                key: samples[:, indices[key]].reshape((repetitions, instances[key], length))
                for key, length in lengths.items()
            }
            results.append(ResultImpl(params=param_resolver, records=records))

        return results



import cirq
import stimcirq


def test_repr():
    r = stimcirq.IIGate()
    assert eval(repr(r), {'stimcirq': stimcirq}) == r


def test_json_serialization():
    r = stimcirq.IIGate()
    c = cirq.Circuit(r(cirq.LineQubit(1), cirq.LineQubit(3)))
    json = cirq.to_json(c)
    c2 = cirq.read_json(json_text=json, resolvers=[*cirq.DEFAULT_RESOLVERS, stimcirq.JSON_RESOLVER])
    assert c == c2


def test_json_backwards_compat_exact():
    raw = stimcirq.IIGate()
    packed = '{\n  "cirq_type": "IIGate"\n}'
    assert cirq.read_json(json_text=packed, resolvers=[*cirq.DEFAULT_RESOLVERS, stimcirq.JSON_RESOLVER]) == raw
    assert cirq.to_json(raw) == packed
